# ------------------------------------------------------------------------
#                  Generic Libraries
# ------------------------------------------------------------------------

import  numpy              as      np
import  matplotlib.pyplot  as      plt
from    scipy.interpolate  import  interp1d
from    os.path            import  dirname, abspath
from    os.path            import  join              as  pjoin

# ------------------------------------------------------------------------
#                  Custom Libraries
# ------------------------------------------------------------------------

from  core.utils  import  lmap, lfilter, reader, qformat

# ------------------------------------------------------------------------
#                  Basic Definitions
# ------------------------------------------------------------------------

root_dir     =  dirname(abspath(__file__))

array        =  np.array
log10        =  np.log10

spacesplit   =  lambda x: lfilter(None, x.split(' '))
parse_float  =  lambda x: lmap(float, spacesplit(x))

logspace   =  lambda a,b,n: 10**np.linspace(log10(a), log10(b), n)

# ------------------------------------------------------------------------
#                  Load Current DNS Data
# ------------------------------------------------------------------------

def mk_dns(args,key):
    
    # load y, u+
    y       =  array(args['data'][key]['y'      ])
    VarPlus =  array(args['data'][key]['profile'])
    
    # plot only one side of the channel
    side        =  y <= (0.5*args['Ly'])
    y, VarPlus  =  y[side], VarPlus[side]
    
    # calculate y+
    yplus       =  y * args['Ret']
    
    # filter low y+ values
    return { 'yplus'  : yplus     ,
             'VarPlus': VarPlus   ,
             'mask'   : yplus > 1.}

# ------------------------------------------------------------------------
#                  Main Class
# ------------------------------------------------------------------------

class Plotter_DUTU_Plus:
    def __init__(self, args):
        
        self.args     =  args
        
        for key in 'UT':
            # load current current dns
            dns     =  mk_dns(args,key)
        
            # write logfile
            self.writer(args,key, dns)
    
    @staticmethod
    def writer(args, key, dns):
        yplus        =  dns['yplus']
        center       =  (yplus > 75.) * (yplus < 150.)
        avg_VarPlus  =  np.average(dns['VarPlus'][center])
        
        with open(args['dat_fname'][key], 'w') as f:
            write    =  lambda s     :  f.write(s+'\n')
            fmt      =  lambda x     :  f'{x:19s}' if (type(x)==str) else f'{x:19.12e}'
            mk_line  =  lambda *all_x:  '    ' + '  '.join(map(fmt,all_x))
            
            lmap(write,args['header'][key])
            write( '\nSummary Statistics:')
            write(f'    Average {key}+ (75 < y+ < 150):  {qformat(avg_VarPlus)}')
            write(f"\nData Series (n = {len(yplus)}):")
            write(mk_line('y+',f'{key}+_dns_fortran'))
            for i in range(len(yplus)):
                 write(mk_line(yplus[i]          ,
                               dns['VarPlus'][i]))