# ------------------------------------------------------------------------
#                  Generic Libraries
# ------------------------------------------------------------------------

from  os       import  listdir
from  os.path  import  abspath, basename, dirname, isfile
from  os.path  import  join     as  pjoin
from  re       import  search   as  re_search

# ------------------------------------------------------------------------
#                  Custom Libraries
# ------------------------------------------------------------------------

from core.utils  import  deepdirname, contains_path, as_namedtuple, spacejoin, lfilter, lmap, import_json, path_as_linux

# ------------------------------------------------------------------------
#                  Basic Definitions
# ------------------------------------------------------------------------

__file__   =  abspath(__file__)


listdir_full =  lambda folder: [pjoin(folder,x) for x in listdir(folder)]

# ------------------------------------------------------------------------
#                  Data Paths
# ------------------------------------------------------------------------

def get_data_dirs():
    root_dir        =  deepdirname(__file__,3)
    output_dir      =  pjoin(root_dir, 'data', 'output')
    input_dir       =  pjoin(root_dir, 'data', 'input' )
    
    return { 'output_dir'      :         output_dir                                  ,
             'surf_input_dir'  :  pjoin( input_dir                , 'surfs'        ) ,
             'case_input_dir'  :  pjoin( input_dir                , 'cases'        ) ,
             'unzipped_dns_dir':  pjoin( deepdirname(__file__,5)  , 'Unzipped_DNS' ) ,
           }

data_dirs  =  as_namedtuple( get_data_dirs() , 'data_dirs' )

# ------------------------------------------------------------------------
#                  Writter
# ------------------------------------------------------------------------

def writer(fname, A):
    
    assert contains_path(fname, data_dirs.output_dir)
    
    with open(fname,'w') as f:
        for x in A:
            f.write( x + '\n' )

# ------------------------------------------------------------------------
#                  Logger
# ------------------------------------------------------------------------

class FPrint:
    def __init__(self, fname, check_path = True, track_all = False):
        
        if check_path:
            assert contains_path(fname, data_dirs.output_dir)
        
        self.fname        =  fname
        self.fwrite       =  open(fname, 'w')
        
        self.__track_all  =  track_all
        self.__A          =  []
        
    def __call__(self,*line):
        
        if line and line[-1] == '__only_to_file__':
            # remove last position
            line = spacejoin(line[:-1])
            # avoid printing
        else:
            line = spacejoin(line)
            print(line)
        
        self.fwrite.write( line + '\n' )
        self.fwrite.flush()
        
        if self.__track_all: self.__A.append(line)
    
    def get_all(self):
        assert self.__track_all
        return self.__A
    
    close   = __del__ = lambda self: self.fwrite.close()
    

# ------------------------------------------------------------------------
#                  Case Info
# ------------------------------------------------------------------------

def Mk_Case(folder):
    folder = path_as_linux(folder)
    
    def __finder(A, end):
        A = lfilter(lambda x: (end in x),A)
        assert len(A) == 1,(A,end)
        return A.pop()
    
    D = {}
    
    assert dirname(folder) == path_as_linux(data_dirs.unzipped_dns_dir)
    
    files            =  lfilter(isfile,listdir_full(folder))
    
    D['subfolder']         =  basename(folder)
    D['tecname'] = tecname = __finder(files,'.tec'                 )
    D['xyzname']           =  __finder(files,'.dat'                )
    D['surf_define']       =  __finder(files,'rough_surf_define.py')
    
    extra_info       =  import_json(__finder(files,'.info'))
    D['Ret']         =  extra_info['Ret'       ]
    D['Pr']          =  extra_info['Pr'        ]
    D['n_xyz']       =  extra_info['n_xyz'     ]
    D['ref_header']  =  extra_info['ref_header']
    D['ref_zone']    =  extra_info['ref_zone'  ]
    
    assert ('_stagTec3d_' in D['tecname']) and ('.tec' in D['tecname'])
    D['tag'] = basename(D['tecname']).replace('stagTec3d_','').replace('.tec','')
    
    D['title'] =  basename(folder)
    
    return as_namedtuple(D,'case_json')
