# %run /home/rafael/ownCloud/workspace/clean_cache.py
# %run /home/rafael/ownCloud/workspace/code/202109_stag_interp_v5/all_interp/code/core/summary_forces.py
# %run /home/rafael/ownCloud/workspace/clean_cache.py
#
# %run C:\Users\rafae\surfdrive\workspace\clean_cache.py
# %run C:\Users\rafae\surfdrive\workspace\code\202109_stag_interp_v5\all_interp\code\core\summary_tec.py
# %run C:\Users\rafae\surfdrive\workspace\clean_cache.py
# ------------------------------------------------------------------------
#                  Generic Libraries
# ------------------------------------------------------------------------

from    sys      import  path   as  sys_path
from    re       import  search as  re_search
from    os       import  listdir
from    os.path  import  basename, abspath, dirname
from    os.path  import  join                        as  pjoin
from    time     import  time
from    math     import  prod

import  torch


# ------------------------------------------------------------------------
#                  Custom Libraries
# ------------------------------------------------------------------------

sys_path.append(dirname(dirname(abspath(__file__))))
from  core.utils     import  ( lmap         ,
                               lfilter      ,
                               listdir_full ,
                               reader       )
from core.io_access  import    data_dirs
sys_path.pop()

# ------------------------------------------------------------------------
#                  Basic Functions
# ------------------------------------------------------------------------

spacesplit = lambda line: lfilter(None,line.split(' '))
max_diff   = lambda A,B : float(torch.abs(A-B).max())
# ------------------------------------------------------------------------
#                  Load Single Tecplot File
# ------------------------------------------------------------------------
#     (a staggered DNS grid is described by multiple files)

class Load_Single_Tec:
    def __init__(self, fname, as_2d = False):
        
        A            =  reader(fname)
        header,zone  =  A[:2]
        A            =  lmap( lambda line: lmap(float,spacesplit(line)) ,
                              A[2:]                                     )
        
        # define order
        order  =  header.split('=')[1].replace('"',' ').replace(',',' ')
        order  =  spacesplit(order)
        
        # convert to torch.tensor
        A      =  torch.tensor(A, dtype = torch.double)
        
        # register all variables
        self.as_2d   =  as_2d
        self.fname   =  fname
        self.header  =  header
        self.zone    =  zone
        self.order   =  order
        self.data    =  { key:A[:,i] for i,key in enumerate(order) }
        
        if as_2d:
            self.__make_all_2d()
    
    def __make_all_2d(self):
        
        inds_2d = self.__find_inds_2d( self.data['X'] , 
                                       self.data['Z'] )
        
        for k,v in self.data.items():
            self.data[k] = v[inds_2d]
    
    def __find_inds_2d(self, X, Z):
        
        # build index maps
        #     (convert xz positions into indexes)
        #     e.g., [0.5 ,1.5, 0.5, ... ] -> [0, 1, 0,...]
        ix, nx1  =  self.__mk_inds(X)
        iz, nz1  =  self.__mk_inds(Z)
        
        # shape 2d
        shape2 = nx1, nz1
        
        # assert shapes are consistent
        #     (floating point errors in "unique" would be detected here)
        assert prod(shape2) == len(X) == len(Z)
        
        # define final array
        #     inds_2d = [[0,1,5,6],[9,3,2]]
        #       (this indicates how elements should be regrouped to create 2d array)
        
        inds_2d = torch.zeros( shape2 , dtype = torch.long )
        
        for k in range(len(X)):
            # inds_2d[a,b] goes at position (ix[k],iz[k])
            inds_2d[ix[k],iz[k]] = k
        
        return inds_2d
    
    def __mk_inds(self, X):
        x1      =  torch.unique(X, sorted=True)
        to_ind  =  { val.item():i for i,val in enumerate(x1) }
        return  [to_ind[val.item()] for val in X], len(x1)

# ------------------------------------------------------------------------
#                  Importer
# ------------------------------------------------------------------------
#     (a staggered DNS grid is described by multiple files)

class SummaryBalances:
    def __init__(self, folder, side, scanned_family = 'quadint_interp2center', fprint = print, write_tec = True):
        self.write_tec       =  write_tec
        self.basic_fields    =  ['X','Y','Z','H','dxdz'] 
        
        self.fprint          =  fprint
        self.folder          =  folder
        self.side            =  side
        self.scanned_family  =  scanned_family
        self.__get_tec()
        self.__scan_log()
        self.__make_summaries()
        
    def __get_tec(self):
        self.tec_objs = {}
        for file in listdir_full(self.folder):
            m = re_search(f'_{self.scanned_family}_{self.side}_'+'(\w+).tec', file)
            if hasattr(m,'groups'):
                m = m.groups()
                assert  (len(m) == 1)  and  (not m[0] in self.tec_objs)
                self.tec_objs[m[0]] = Load_Single_Tec(file)
        assert set(self.tec_objs.keys()) == set('UVWPT')
    
    def __scan_log(self):
        self.logfile = lfilter(lambda x: '.log' in basename(x),
                               listdir_full(self.folder)      )
        assert len(self.logfile) == 1
        self.logfile  =  self.logfile[0]
        A             =  reader(self.logfile)
        
        all_tags = lfilter(lambda tag: not tag in self.basic_fields                                      ,
                           sorted(set( tag for obj in self.tec_objs.values() for tag in obj.data.keys() )))
        
        self.tags_to_idir_type = D = {}
        for line in A:
            for tag in all_tags:
                m = re_search(f'\({tag}\)' + ' \((\w+)\) \((\w+)\):', line)
                if hasattr(m,'groups'):
                    idir, typ =  m.groups()
                    idir      =  int(idir)
                    # non-overwrite policy
                    assert (not tag in D) or (D[tag] == (idir, typ))
                    D[tag] = (idir, typ)
    
    def __make_summaries(self):
        self.all_fields  =  all_fields  =  {nature: [{} for _ in range(3)] for nature in ['F','Qdot']}
        
        get_nature       =  {typ:'F' for typ in 'UVWP'}
        get_nature['T']  =  'Qdot'
        
        for typ,obj in self.tec_objs.items():
            assert not obj.as_2d
            fields = all_fields[get_nature[typ]]
            for key,arr in obj.data.items():
            
                if key in self.basic_fields:
                    for idir2 in range(3):
                        D = fields[idir2]
                        if key in D:
                            assert max_diff(arr,D[key]) < 1e-10, f'ERROR: (key: {key}) (tol: {max_diff(arr,D[key])}) (typ: {typ}, nature: {nature})'
                        D[key] = arr
                
                else:
                    idir2, typ2 = self.tags_to_idir_type[key]
                    assert (typ == typ2) and (not key in fields[idir2])
                    fields[idir2][key] = arr
        # merge Qdot data (temp gradients were split by component)
        all_fields['Qdot'] = [ {k:v for D in all_fields['Qdot'] for k,v in D.items()} ]
        
        self.fprint('\n-----------------------------------------------------\n    TecPlot File: Net Balance\n-----------------------------------------------------\n')
        self.fprint(f'Scanned Family: {self.scanned_family}')
        self.fprint(f'Side          : {self.side}')
        self.fprint(f'Folder        : {self.folder}')
        self.fprint(f'Balances      :')
        
        for nature,fields in all_fields.items():
            for i,D in enumerate(fields):
                self.__mk_net_balance(         D)
                self.__mk_tecpost    (i,nature,D)
                self.fprint(f'    {nature:4s}[{i}]: {D["net_balance"].sum().item():8.3f}')
    
    def __mk_net_balance(self,D):
        s = 0.
        for key,arr in D.items():
            if key in self.tags_to_idir_type:
                s += arr
            else:
                assert key in self.basic_fields
        D['net_balance'] = s
    
    def __mk_tecpost(self,i,nature,D):
        if not self.write_tec:
            return
        
        keys   =  self.basic_fields + ['net_balance']
        keys  +=  sorted(set(D.keys())-set(keys))
        
        header = 'VARIABLES ='+ ', '.join(f'"{k}"' for k in keys)
        zone   = list(self.tec_objs.values())[0].zone
        
        fname_w = self.logfile.replace('.log',f'_{self.scanned_family}_{nature}{i}_net.tec')
        L      = set()
        with open(fname_w,'w') as f:
            write = lambda s: f.write(str(s)+'\n')
            write(header)
            write(zone)
            L |= set(map(len,D.values()))
            assert len(L) == 1
            for i in range(list(L)[0]):
                write(' '.join(map(str,[D[k][i].item() for k in keys])))


# ------------------------------------------------------------------------
#                  Direct Runner
# ------------------------------------------------------------------------

def print_compare(s,R, fprint = print):
    fprint('')
    fprint(f'File 1: {basename(R.logfile)}')
    fprint(f'File 2: {basename(s.logfile)}')
    fabs = lambda arr: torch.abs (arr)
    favg = lambda arr: torch.mean(arr)    .item()
    fmax = lambda arr:            arr.max().item()
    for nature,fields_1 in R.all_fields.items():
        fields_2 = s.all_fields[nature]
        for i in range(len(fields_1)):
           n1        =  fields_1[i]["net_balance"]
           n2        =  fields_2[i]["net_balance"]
           L1        =  favg(fabs(n1-n2))/favg(fabs(n1))
           diff_max  =  fmax(fabs(n1-n2))/favg(fabs(n1))
           fprint(f'    {nature:4s}[{i}] (side = {R.side:6s})(b1 = {n1.sum().item():8.3f}) (b2 = {n2.sum().item():8.3f}) (L1 (avg/avg) = {(L1*100):.6f} %) (diff_max (max/avg) = {(diff_max*100):.6f} %)')

# ------------------------------------------------------------------------
#                  Direct Runner
# ------------------------------------------------------------------------

if __name__ == '__main__':
    
    is_ref  =  lambda folder:  basename(folder)[:4] == 'P400'
    for     side   in ['top','bottom']:
        A = []
        for folder in sorted(filter( lambda s: basename(s)[0] == 'P'    ,
                                     listdir_full(data_dirs.output_dir) )):
            s = SummaryBalances(folder, side, write_tec = False)
            if is_ref(folder):
                R = s
            else:
                A.append(s)
        for s in A:
            print_compare(s,R)