# ------------------------------------------------------------------------
#                  Generic Libraries
# ------------------------------------------------------------------------

from  collections  import  namedtuple
from  os.path      import  dirname
from  time         import  time
from  datetime     import  datetime
from  math         import  prod
from  psutil       import  Process  as  psutil_Process
from  os           import  getpid, listdir
from  os.path      import  basename, dirname, abspath
from  os.path      import  join     as  pjoin

# ------------------------------------------------------------------------
#                  Basic Definitions
# ------------------------------------------------------------------------

lmap              =  lambda f,x:  list(map   (f,x))
lfilter           =  lambda f,x:  list(filter(f,x))

deepdirname       =  lambda x,n:  x if n<1 else deepdirname(dirname(x),n-1)
spacejoin         =  lambda x  :  ' '.join(map(str,x))

get_ram_usage_gb  =  lambda: psutil_Process(getpid()).memory_info().rss/1e9

as_item           =  lambda x: x.item()
transpose         =  lambda A: lmap(list,zip(*A))

listdir_full      =  lambda x: [pjoin(x,y) for y in listdir(x)]

def pop1(A):
    assert (len(A) == 1) and (type(A)==list)
    return A[0]

# ------------------------------------------------------------------------
#                  Quick Test (Transpose)
# ------------------------------------------------------------------------

# ensure that transpose works properly
assert transpose( [[  1 ,  2 ,  3 ],
                   [  4 ,  5 ,  6 ],
                   [  7 ,  8 ,  9 ],
                   [ 10 , 11 , 12 ]] )  ==  [[1,4,7,10],[2,5,8,11],[3,6,9,12]]

# ------------------------------------------------------------------------
#                  Replace Username
# ------------------------------------------------------------------------

def path_as_linux(x):
    cWin         = r' \ '.replace(' ','')
    cLin         = r' / '.replace(' ','')
    # iters, Nmax  =  0, max(len(x)*10, int(1e4))
    x_old        =  None
    while x_old != x:
        x_old = x
        x = x.replace(cWin    , cLin)
        x = x.replace(cLin+cLin,cLin)
        # iters += 1
        # assert iters < Nmax
    return x

def splitpath(x):
    x     = path_as_linux(x)
    cLin  =  r' / '.replace(' ','')
    return lfilter(None, x.split(cLin))

def split_upto(x, tag):
    x = path_as_linux(x)
    assert tag in splitpath(x)
    A            = []
    # iters, Nmax  =  0, max(len(x)*10,int(1e4))
    while basename(x) != tag:
        A.append(basename(x))
        x = dirname(x)
        # iters += 1
        # assert iters < Nmax
    assert basename(x) == tag
    return x, A[::-1]

def fix_workspace_path(x, tag = 'workspace'):
    x = path_as_linux(x)
    return path_as_linux(pjoin( split_upto(path_as_linux(abspath(__file__)), tag)[0] ,
                               *split_upto(x                               , tag)[1] ))

# ------------------------------------------------------------------------
#                  Time Stamps
# ------------------------------------------------------------------------

tstamp     =  lambda   :  datetime.now().strftime("%Y%m%d_%H%M%S")
full_time  =  lambda   :  datetime.now().strftime("%H:%M:%S, %d/%m/%Y")
format_dt  =  lambda x :  "%02d:%02d:%02d [hh:mm:ss]" % (x//3600, (x%3600)//60, x%60)

# ------------------------------------------------------------------------
#                  Format Numbers
# ------------------------------------------------------------------------

iformat    =  lambda x: f'  {x:9d}'     if not hasattr(x,'__iter__') else f'[ {(x[0]  ):9d}    ,  {(x[1]  ):9d}    ,  {(x[2]  ):9d}   ]'
qformat    =  lambda x: f'  {x:9.5f}'   if not hasattr(x,'__iter__') else f'[ {(x[0]):9.5f}    ,  {(x[1]):9.5f}    ,  {(x[2]):9.5f}   ]'
performat  =  lambda x: f'  {x:9.5f} %' if not hasattr(x,'__iter__') else f'[ {(x[0]):9.5f} %  ,  {(x[1]):9.5f} %  ,  {(x[2]):9.5f} % ]'

# ------------------------------------------------------------------------
#                  Format Conversions
# ------------------------------------------------------------------------

def as_namedtuple(D, name):
    keys,values = [],[]
    for k,v in D.items():
        keys  .append(k)
        values.append(v)
    return namedtuple(name,keys)(*values)

def as_list(x, dtype):
    if hasattr(x,'__iter__') and ((prod(x.shape)>1) if hasattr(x,'shape') else True):
        return [ as_list(c,dtype) for c in x ]
    else:
        return dtype(x)

# ------------------------------------------------------------------------
#                  Path Verification
# ------------------------------------------------------------------------

def contains_path(x,p):
    x = path_as_linux(x)
    p = path_as_linux(p)
    # iters, Nmax  =  0, max(len(x)*10,int(1e4))
    while x:
        if x == p:
            return True
        x = dirname(x)
        # iters += 1
        # assert iters < Nmax
    raise Exception(f'Path (x = {x}) is not located within (root = {p})')

# ------------------------------------------------------------------------
#                  Reader Functions
# ------------------------------------------------------------------------

def reader(fname, clean = False):
    
    A = []
    with open(fname,'r') as f:
        for x in f:
            A.append(x.rstrip('\n'))
    
    if clean:
        A = lfilter(None,A)
    
    return A

import_json  =  lambda fname: eval(' '.join(reader(fname)))
